from . import Engine, register_engine
from .dadi_moments_common import DadiOrMomentsEngine
from ..models import DemographicModel, CustomDemographicModel, Epoch, Split
from ..utils import DynamicVariable
from .. import SFSDataHolder, VCFDataHolder
import numpy as np


class MomentsEngine(DadiOrMomentsEngine):
    """
    Engine for using :py:mod:`moments` for demographic inference.

    Citation of :py:mod:`moments`:

    Jouganous, J., Long, W., Ragsdale, A. P., & Gravel, S. (2017).
    Inferring the joint demographic history of multiple populations:
    beyond the diffusion approximation. Genetics, 206(3), 1549-1567.
    """

    id = 'moments'  #:
    import moments as base_module
    supported_data = [SFSDataHolder]  # , VCFDataHolder]  #:
    inner_data_type = base_module.Spectrum  #:
    default_dt_fac = 0.01  #:

    @staticmethod
    def _get_kwargs(event, var2value):
        """
        Builds kwargs for moments.Spectrum.integrate functions.

        :param event: build for this event
        :type event: event.Epoch
        :param var2value: dictionary {variable: value}, it is required because
            the dynamics values should be fixed.
        """
        ret_dict = {'tf': event.time_arg}
        if event.dyn_args is not None:
            ret_dict['Npop'] = list()
            for i in range(event.n_pop):
                dyn_arg = event.dyn_args[i]
                dyn = MomentsEngine.get_value_from_var2value(var2value,
                                                             dyn_arg)
                if dyn == 'Sud':
                    ret_dict['Npop'].append(event.size_args[i])
                else:
                    ret_dict['Npop'].append('nu%d_func' % (i+1))
        else:
            ret_dict['Npop'] = event.size_args

        if event.mig_args is not None:
            ret_dict['m'] = np.zeros(shape=(event.n_pop, event.n_pop),
                                     dtype=object)
            for i in range(event.n_pop):
                for j in range(event.n_pop):
                    if i == j:
                        continue
                    ret_dict['m'][i, j] = event.mig_args[i][j]
        if event.sel_args is not None:
            ret_dict['gamma'] = event.sel_args
        if event.dom_args is not None:
            ret_dict['h'] = event.dom_args
        return ret_dict

    def _inner_func(self, values, ns, dt_fac=0.01):
        """
        Simulates expected SFS for proposed values of variables.

        :param values: values of variables
        :param ns: sample sizes of simulated SFS
        :param dt_fac: step size for numerical solution
        """
        if isinstance(self.model, CustomDemographicModel):
            return self.model.function(values, ns)

        var2value = self.model.var2value(values)
        moments = self.base_module

        sts = moments.LinearSystem_1D.steady_state_1D(np.sum(ns))
        fs = moments.Spectrum(sts)

        ns_on_splits = [list(ns)]
        for ind, event in enumerate(reversed(self.model.events)):
            if isinstance(event, Split):
                p = event.pop_to_div
                ns_on_splits.append(ns_on_splits[-1][:-1])
                ns_on_splits[-1][p] += ns_on_splits[-2][-1]
        ns_on_splits = list(reversed(ns_on_splits))[1:]
        n_split = 0

        addit_values = {}
        for ind, event in enumerate(self.model.events):
            if isinstance(event, Epoch):
                if event.dyn_args is not None:
                    for i in range(event.n_pop):
                        dyn_arg = event.dyn_args[i]
                        value = self.get_value_from_var2value(var2value,
                                                              dyn_arg)
                        if value != 'Sud':
                            func = DynamicVariable.get_func_from_value(value)
                            y1 = self.get_value_from_var2value(
                                var2value, event.init_size_args[i])
                            y2 = self.get_value_from_var2value(
                                var2value, event.size_args[i])
                            x_diff = self.get_value_from_var2value(
                                var2value, event.time_arg)
                            addit_values['nu%d_func' % (i+1)] = func(
                                y1=y1,
                                y2=y2,
                                x_diff=x_diff)
                kwargs_with_vars = self._get_kwargs(event, var2value)
                kwargs = {}
                for x, y in kwargs_with_vars.items():
                    if x == 'm' and isinstance(y, np.ndarray):  # migration
                        l_s = np.array(y, dtype=object)
                        for i in range(y.shape[0]):
                            for j in range(y.shape[1]):
                                yel = y[i][j]
                                l_s[i][j] = self.get_value_from_var2value(
                                    var2value, yel)
                                l_s[i][j] = addit_values.get(l_s[i][j],
                                                             l_s[i][j])
                        kwargs[x] = l_s
                    elif x == 'Npop':  # sizes
                        n_pop_list = list()
                        all_sudden = True
                        for y1 in y:
                            n_pop_list.append(self.get_value_from_var2value(
                                var2value, y1))
                            if n_pop_list[-1] in addit_values:
                                all_sudden = False
                        if not all_sudden:
                            kwargs[x] = lambda t: [addit_values[el](t)
                                                   if el in addit_values
                                                   else el
                                                   for el in list(n_pop_list)]
                        else:
                            kwargs[x] = n_pop_list
                    elif x in ['gamma', 'h']:
                        l_args = list()
                        for i in range(len(y)):
                            l_args.append(self.get_value_from_var2value(
                                    var2value, y[i]))
                        kwargs[x] = l_args
                    else:  # time
                        kwargs[x] = self.get_value_from_var2value(var2value, y)
                        kwargs[x] = addit_values.get(kwargs[x], kwargs[x])
                fs.integrate(dt_fac=dt_fac, **kwargs)
            elif isinstance(event, Split):
                ns_split = (ns_on_splits[n_split][event.pop_to_div],
                            ns_on_splits[n_split][-1])
                if event.n_pop == 1:
                    fs = moments.Manips.split_1D_to_2D(fs, *ns_split)
                else:
                    func_name = "split_%dD_to_%dD_%d" % (
                        event.n_pop, event.n_pop + 1, event.pop_to_div + 1)
                    fs = getattr(moments.Manips, func_name)(fs, *ns_split)
                n_split += 1
        return fs

    def draw_schematic_model_plot(self, values, save_file=None,
                                  fig_title="Demographic Model from GADMA",
                                  nref=None, gen_time=1,
                                  gen_time_units="Generations"):
        """
        Draws schematic plot of the model with values.
        See moments manual for more information.

        :param values: Values of the model parameters, it could be list of
                       values or dictionary {variable name: value}.
        :type values: list or dict
        :param save_file: File to save picture. If None then picture will be
                          displayed to the screen.
        :type save_file: str
        :param fig_title: Title of the figure.
        :type fig_title: str
        :param nref: An ancestral population size. If None then parameters
                     will be drawn in genetic units.
        :type nref: int
        :param gen_type: Time of one generation. Should be in units of
                         ``gen_time_units``.
        :type gen_type: float
        :param gen_time_units: Units of `gen_type`. For example, it
                               could be Years, Generations, Thousand Years and
                               so on.
        """
        moments = self.base_module
        # From moments docs about ns:
        # List of sample sizes to be passed as the ns argument to model_func.
        # Actual values do not matter, as long as the dimensionality is
        # correct. So we take small size for fast drawing.
        ns = [4 for _ in range(len(self.data.sample_sizes))]
        plot_mod = moments.ModelPlot.generate_model(self._inner_func,
                                                    values, ns)
        draw_scale = nref is not None
        show = save_file is None
        if nref is not None:
            nref = int(nref)
        moments.ModelPlot.plot_model(plot_mod,
                                     save_file=save_file,
                                     show=show,
                                     fig_title=fig_title,
                                     draw_scale=draw_scale,
                                     pop_labels=self.data.pop_ids,
                                     nref=nref,
                                     gen_time=gen_time,
                                     gen_time_units=gen_time_units,
                                     reverse_timeline=True)

    def simulate(self, values, ns, dt_fac=default_dt_fac):
        """
        Returns simulated expected SFS for :attr:`demographic_model` with
        values as parameters.

        :param values: Values of demographic model parameters.
        :param ns: sample sizes of the simulated SFS.
        """
        moments = self.base_module
        model = self._inner_func(values, ns, dt_fac)
        return model

    def get_theta(self, values, dt_fac=default_dt_fac):
        return super(MomentsEngine, self).get_theta(values, dt_fac)

    def evaluate(self, values, dt_fac=default_dt_fac):
        """
        Simulates SFS from values and evaluate log likelihood between
        simulated SFS and observed SFS.
        """
        try:
            val = super(MomentsEngine, self).evaluate(values, dt_fac)
            return val
        except RuntimeError as e:
            if str(e) == "Factor is exactly singular":
                return None
            raise e

    def generate_code(self, values, filename=None, dt_fac=default_dt_fac,
                      nanc=None, gen_time=None, gen_time_units=None):
        return super(MomentsEngine, self).generate_code(values, filename,
                                                        dt_fac, nanc, gen_time,
                                                        gen_time_units)


register_engine(MomentsEngine)



import copy
import numpy as np
from functools import wraps
import sys
import os

from ..utils import Variable, ContinuousVariable
from ..utils import fix_args, cache_func, nan_fval_to_inf, WeightedMetaArray
from ..utils import ensure_file_existence, check_file_existence,\
                    variables_values_repr
from ..utils import logarithm_transform, exponent_transform, ident_transform
import pickle
import warnings


class Optimizer(object):
    """
    Base class for optimizer. The most important methods:
    :meth:`gadma.optimizers.Optimizer.evaluate` and
    :meth:`gadma.optimizers.Optimizer.optimize`.

    To create new class for optimizer one should at least implement
    :meth:`gadma.optimizers.Optimizer.valid_restore_file` and
    :meth:`gadma.optimizers.Optimizer.optimize`

    :param log_transform: If True then all parameters are optimized in log
                          scale.
    :type log_transform: bool
    :param maximize: If True then maximization of target function is
                     performed.
    :type maximize: bool
    """
    def __init__(self, log_transform=False, maximize=False):
        self.log_transform = log_transform
        self.maximize = maximize

    @property
    def log_transform(self):
        return self._log_trasform

    @log_transform.setter
    def log_transform(self, log_transform):
        self._log_trasform = log_transform
        if log_transform:
            self.transform = logarithm_transform
            self.inv_transform = exponent_transform
        else:
            self.transform = ident_transform
            self.inv_transform = ident_transform

    @property
    def sign(self):
        """
        Returns -1 if maximization and 1 if minimization of target function.
        """
        return -1 if self.maximize else 1

    def prepare_callback(self, callback):
        """
        Wraps callback function for usage inside optimizer. It transforms x
        according to `log_transform` and multiply y to -1 if `maximize` is
        True.

        :param callback: function of callback that has the following notation:
                         `callback(x, y)`, where x is value of parameters and
                         y is the value of target function on `x`.
        """
        @wraps(callback)
        def wrapper(x, y):
            x_tr = self.inv_transform(x)
            if isinstance(x, WeightedMetaArray):
                x_tr = WeightedMetaArray(x_tr)
                x_tr.metadata = x.metadata
            return callback(x_tr, self.sign * y)
        return wrapper

    def write_report(self, n_iter, variables, x, y, report_file):
        """
        Writes report of optimizer to file or stdout.

        :param n_iter: Number of iteration of optimization.
        :param variables: list of variables which values are optimized.
        :param x: Values of variables.
        :param y: Value of target function on `x`.
        :param report_file: filename to write report. If None then report is
                            printed to stdout.
        """
        if report_file:
            stream = open(report_file, 'a')
        else:
            stream = sys.stdout
        x_repr = variables_values_repr(variables, x)
        metainfo = ''
        if hasattr(x, 'metadata'):
            metainfo = x.metadata
        print(n_iter, y, x_repr, metainfo, sep='\t', file=stream)
        if report_file:
            stream.close()

    def wrap_for_report(self, f, variables, verbose, report_file):
        """
        Wraps function `f` for automatically report output. When function is
        called report is saved to `report_file` every `verbose` call.

        :param f: Function to wrap.
        :param variables: Variables of function `f`.
        :param verbose: Verbosity level.
        :param report_file: Filename to save report.
        """
        @wraps(f)
        def wrapper(x):
            wrapper._counter += 1
            y = f(x)
            if (verbose > 0) and (wrapper._counter % verbose == 0):
                self.write_report(wrapper._counter, variables, x, y,
                                  report_file)
            return y
        if report_file:
            ensure_file_existence(report_file)
        wrapper._counter = 0
        if verbose == 0:
            return f
        return wrapper

    def evaluate(self, f, x, args=(), linear_constrain=None):
        """
        Evaluates function `f` on values `x`.

        :param f: Target function.
        :param x: Value of parameters of `f`.
        :param args: Other arguments of f. `f(x, args)`.
        :param linear_constrain: Linear constrain on `x`.
        """
        x_tr = self.inv_transform(x)
        if linear_constrain is not None:
            if not linear_constrain.fits(x_tr):
                x_tr, success = linear_constrain.try_to_transform(x_tr)
                if not success:
                    warnings.warn(f"HERE IS A LITTLE PROBLEM. PLEASE CHECK "
                                  f"IT: {x}, {x_tr}")
                    return self.sign * np.inf
        y = f(x_tr, *args)
        return self.sign * y

    def prepare_f_for_opt(self, f, args=(), cache=True):
        """
        Prepares `f` for usage in optimizer. It should be transformed
        according to `log_transform` and `maximize`. Arguments are fixed and
        it could be cached or not.

        :param f: Target function to work with.
        :param args: Arguments of the function. `f(x, args)`.
        :type args: tuple
        :param cache: If True then function is cached.
        :type cache: bool.
        """
        assert isinstance(cache, bool)
        # Fix args
        f_wrapped = fix_args(f, *args)
        f_wrapped = nan_fval_to_inf(f_wrapped)
        if cache:
            f_wrapped = cache_func(f_wrapped)
        return f_wrapped

    def check_variables(self, variables):
        """
        Checks that all `variables` are instances of
        :class:`gadma.utils.Variable` class.
        """
        for var in variables:
            assert isinstance(var, Variable)

    def save(self, info, save_file):
        """
        Save some information into file. Is supposed to save info during
        optimization in order to restore it.

        :param info: Information to dump.
        :param save_file: File to save information.

        :note: if save_file is None then nothing will be done. In base class\
               method just dumps `info` to `save_file` with `pickle`.
        """
        if save_file is None:
            return
        if hasattr(self, 'id'):
            if (check_file_existence(save_file) and
                    os.path.getsize(save_file) > 0):
                with open(save_file, 'rb') as fl:
                    d = pickle.load(fl)
                if not isinstance(d, dict):
                    d = {}
            else:
                d = {}
            d[self.id] = copy.copy(info)
            info = d
        with open(save_file, 'wb') as fl:
            pickle.dump(info, fl)

    def valid_restore_file(self, save_file):
        """
        Checks that `save_file` contains valid information and it could be
        restored from it.

        :param save_file: File to check.
        """
        raise NotImplementedError

    def load(self, save_file):
        """
        Loads information that was saved by :meth:`save` method.

        :param save_file: File to restore information from.

        :note: In base class method just loads from `save_file` with pickle.
        """
        with open(save_file, 'rb') as fl:
            info = pickle.load(fl)
        if hasattr(self, 'id') and isinstance(info, dict):
            return info[self.id]
        return info

    def optimize(f, variables, args=(), options={}, linear_constrain=None,
                 maxiter=None, maxeval=None, verbose=0, callback=None,
                 report_file=None, eval_file=None, save_file=None,
                 restore_file=None, restore_points_only=False,
                 restore_x_transform=None):
        """
        Run optimization for target function.

        :param f: Target function to optimize.
        :type f: func
        :param variables: List of variables which values are optimized.
        :type variables: :class:`gadma.utils.VariablePool`
        :param args: Additional arguments of target function.
        :type args: tuple
        :param options: Additional options kwargs for optimization.
        :type options: dict
        :param linear_constrain: Linear constrain on optimized variables. It is
                                 optional argument. Could be missed in
                                 unconstrained optimizers.
        :type linear_constrain: :class:`gadma.optimizers.LinearConstrain`
        :param maxiter: Maximum number of iterations to run. If None then run
                        until converge.
        :type maxiter: int
        :param maxeval: Maximum number of evaluations to run. If None then run
                        until converge.
        :type maxeval: int
        :param verbose: Verbosity of the output. If 0 then no reports.
        :type verbose: int
        :param callback: Callback to run after each iteration of optimization.
                         Should be called as `callback(x, y)`
        :type callback: function
        :param report_file: File to save report. Check option `verbose`.
        :type report_file: str
        :param eval_file: File to save all evaluations of the function `f`.
        :type eval_file: str
        :param save_file: File to save information during optimization for its
                          reconstruction.
        :type save_file: str
        :param restore_file: File to restore previous run.
        :type restore_file: str
        :param restore_points_only: Restore point/points from previous run and
                                    run optimization from them once more. If
                                    False then previous run will be resumed.
        :type restore_points_only: bool
        :param restore_x_transform: Restore points but transform them before
                                    usage in this run.
        :type restore_x_transform: function
        """
        raise NotImplementedError


class ContinuousOptimizer(Optimizer):
    """
    Base class for optimization of continuous variables.
    """
    def check_variables(self, variables):
        """
        Returns True if all variables are instances of
        :class:`gadma.utils.ContinousVariable` class.
        """
        for var in variables:
            assert isinstance(var, ContinuousVariable)
        super(ContinuousOptimizer, self).check_variables(variables)


class UnconstrainedOptimizer(ContinuousOptimizer):
    """
    Base class for unconstrained optimization, i.e. when values of variables
    have no bounds.
    """
    def check_variables(self, variables):
        """
        Returns True if all variables have domain of [-inf, inf].
        """
        super(UnconstrainedOptimizer, self).check_variables(variables)
        for var in variables:
            assert np.allclose(var.domain, np.array([-np.inf, np.inf]))


class ConstrainedOptimizer(ContinuousOptimizer):
    """
    Base class for constrained optimization, i.e. when values of variables
    have some bounds.
    """
    def check_variables(self, variables):
        """
        Returns True if all variables have constrained domain.
        """
        super(ConstrainedOptimizer, self).check_variables(variables)
        for var in variables:
            assert np.all(var.domain != np.array([-np.inf, np.inf]))



from ..models import CustomDemographicModel, Epoch, Split, BinaryOperation
from ..utils import DiscreteVariable, DynamicVariable, Variable
from .dadi_generator import FUNCTION_NAME, _print_p0, _print_main,\
                            _print_load_data
import numpy as np
import copy
import sys


def _print_moments_func(model, values, dt_fac):
    """
    Returns string with function of demographic model for :py:mod:`moments`.
    Parameter `values` is needed to fix dynamics of populations.

    :param model: Demographic model.
    :type model: :class:`gadma.models.model.Model`
    :param values: List of values for parameters of model.
    """
    import moments
    from ..engines import MomentsEngine  # to avoid cross import

    if isinstance(model, CustomDemographicModel):
        ret_str = "import importlib.util\n\n"
        ret_str += "spec = importlib.util.spec_from_file_location('module',"\
                   f" '{sys.modules[model.function.__module__].__file__}')\n"
        ret_str += "module = importlib.util.module_from_spec(spec)\n"
        ret_str += "spec.loader.exec_module(module)\n"
        ret_str += f"{FUNCTION_NAME} = module.model_func\n\n"
        return ret_str

    var2value = model.var2value(values)  # OrderedDict

    f_vars = [x for x in var2value if not isinstance(x, DiscreteVariable)]
    ret_str = f"def {FUNCTION_NAME}(params, ns):\n"
    ret_str += "\t%s = params\n" % ", ".join([x.name for x in f_vars])
    ret_str += "\tsts = moments.LinearSystem_1D.steady_state_1D"\
               "(np.sum(ns))\n"\
               "\tfs = moments.Spectrum(sts)\n"

    ns_on_splits = [[f'ns[{i}]'
                     for i in range(model.number_of_populations())]]
    for ind, event in enumerate(reversed(model.events)):
        if isinstance(event, Split):
            p = event.pop_to_div
            ns_on_splits.append(copy.copy(ns_on_splits[-1][:-1]))
            ns_on_splits[-1][p] += " + " + ns_on_splits[-2][-1]
    ns_on_splits = list(reversed(ns_on_splits))[1:]
    n_split = 0

    for ind, event in enumerate(model.events):
        if event.__class__ is Epoch:
            all_sudden = True
            func_names = []
            if event.dyn_args is not None:
                for i in range(event.n_pop):
                    dyn_arg = event.dyn_args[i]
                    value = var2value.get(dyn_arg, dyn_arg)
                    if value != 'Sud':
                        all_sudden = False
                        func = DynamicVariable._help_dict[value]
                        y1 = event.init_size_args[i]
                        y2 = event.size_args[i]
                        x_diff = event.time_arg
                        if isinstance(y1, Variable):
                            y1 = y1.name
                        elif isinstance(y1, BinaryOperation):
                            y1 = f"({y1.name})"
                        if isinstance(y2, Variable):
                            y2 = y2.name
                        elif isinstance(y2, BinaryOperation):
                            y2 = f"({y2.name})"
                        if isinstance(x_diff, Variable):
                            x_diff = x_diff.name
                        elif isinstance(x_diff, BinaryOperation):
                            x_diff = f"({x_diff.name})"
                        funcstr = func.func_str(y1, y2, x_diff)
                        ret_str += "\tnu%d_func = %s\n" % (i+1, funcstr)
                        func_names.append("nu%d_func" % (i+1))
            kwargs_with_vars = MomentsEngine._get_kwargs(event, var2value)
            kwargs = {}
            for x, y in kwargs_with_vars.items():
                if isinstance(y, np.ndarray):  # migration
                    l_s = "m = np.array(["
                    rows = []
                    for i in range(y.shape[0]):
                        elems = []
                        for j in range(y.shape[1]):
                            if i == j:
                                elems.append("0")
                                continue
                            yel = y[i][j]
                            if isinstance(yel, Variable):
                                elems.append(yel.name)
                            else:
                                elems.append(str(yel))
                        rows.append(f"[{', '.join(elems)}]")
                    ret_str += f"\tmigs = np.array([{', '.join(rows)}])\n"
                    kwargs[x] = 'migs'
                elif isinstance(y, list):  # pop. sizes, selection, dynamics
                    varnames = [var.name
                                if isinstance(var, (Variable,
                                                    BinaryOperation))
                                else str(var) for var in y]
                    if x == 'Npop':  # pop size
                        if all_sudden:
                            kwargs[x] = f"[{', '.join(varnames)}]"
                        else:
                            strs = []
                            for y1 in y:
                                if isinstance(y1, (Variable,
                                                   BinaryOperation)):
                                    strs.append(y1.name)
                                else:
                                    if y1 in func_names:
                                        strs.append(y1 + "(t)")
                                    else:
                                        strs.append(y1)
                            kwargs[x] = f"lambda t: [{', '.join(strs)}]"
                    else:
                        kwargs[x] = f"[{','.join(varnames)}]"
                else:  # time
                    kwargs[x] = y.name if isinstance(
                        y, (Variable, BinaryOperation)) else y

            kwargs = [f"{key}={value}" for key, value in kwargs.items()]
            ret_str += f"\tfs.integrate({', '.join(kwargs)}, "\
                       f"dt_fac={dt_fac})\n"
        elif event.__class__ is Split:
            ns_split = (ns_on_splits[n_split][event.pop_to_div],
                        ns_on_splits[n_split][-1])
            if event.n_pop == 1:
                ret_str += "\tfs = moments.Manips.split_1D_to_2D"
            else:
                ret_str += "\tfs = moments.Manips.split_%dD_to_%dD_%d" % (
                           event.n_pop, event.n_pop + 1, event.pop_to_div + 1)

            ret_str += f"(fs, {', '.join(ns_split)})\n"
            n_split += 1
    ret_str += "\treturn fs\n"
    return ret_str


def _is_fs_via_moments(data_holder):
    """
    Check that data is allele frequency spectrum saved for moments.

    :param data_holder: Data holder with data.
    :type data_holder: :class:`gadma.data.data_holder.DataHolder`
    """

    import moments
    try:
        data = moments.Spectrum.from_file(data_holder.filename)
        return data
    except Exception as e:
        return None


def _print_moments_load_data(data_holder):
    is_fs = _is_fs_via_moments(data_holder)
    return _print_load_data(data_holder, is_fs, mode='moments')


def _print_moments_simulation():
    ret_str = f"model = {FUNCTION_NAME}(p0, ns)\n"
    return ret_str


def _print_model_plotting(engine, nanc, gen_time, gen_time_units):
    ret_str = "\n\nplot_ns = [4 for _ in ns]  # small sizes for fast drawing\n"
    ret_str += f"gen_mod = moments.ModelPlot.generate_model({FUNCTION_NAME},\n"
    ret_str += "                                           p0, plot_ns)\n"

    filename = 'model_from_GADMA.png'
    pop_ids = list(engine.data_holder.population_labels)
    draw_scale = nanc is not None
    units = gen_time_units
    if nanc is not None:
        nanc = int(nanc)
    if gen_time is None:
        gen_time = 1.0
    fig_title = "Demographic model from GADMA"
    ret_str += f"moments.ModelPlot.plot_model(gen_mod,\n"
    ret_str += f"                             save_file='{filename}',\n"
    ret_str += f"                             fig_title='{fig_title}',\n"
    ret_str += f"                             draw_scale={draw_scale},\n"
    ret_str += f"                             pop_labels={pop_ids},\n"
    ret_str += f"                             nref={nanc},\n"
    ret_str += f"                             gen_time={gen_time},\n"
    ret_str += f"                             gen_time_units='{units}',\n"
    ret_str += f"                             reverse_timeline=True)"
    return ret_str


def _print_moments_main(engine, values, nanc, gen_time, gen_time_units):
    ret_str = _print_p0(values)
    ret_str += _print_moments_simulation()
    ret_str += _print_main(engine, values, mode='moments', nanc=nanc)
    ret_str += _print_model_plotting(engine, nanc, gen_time, gen_time_units)
    return ret_str


def print_moments_code(engine, values, dt_fac, filename,
                       nanc=None, gen_time=None, gen_time_units=None):
    """
    Generates code for `moments` to file. Code have function of demographic
    model that simulates AFS and main part where simulation takes place as well
    as calculation of log-likelihood.

    :param engine: Engine that was used with data and model.
    :param values: Value of model parameters.
    :param dt_fac: Grid step for moments.
    :param filename: File to save generated code.
    :param nanc: Size of ancestral population. Is used when other engine was
                 used for inference.
    :param gen_time: Time of one generation in units of ``gen_time_units``.
    :param gen_time_units: Units of time. String.

    """

    ret_str = "import moments\nimport numpy as np\n\n"
    ret_str += _print_moments_func(engine.model, values, dt_fac)
    ret_str += "\n"
    ret_str += _print_moments_load_data(engine.data_holder)
    ret_str += "ns = data.sample_sizes\n\n"

    values = [val for var, val in engine.model.var2value(values).items()
              if not isinstance(var, DiscreteVariable)]
    ret_str += _print_moments_main(engine, values, nanc,
                                   gen_time, gen_time_units)
    if filename is None:
        return ret_str
    with open(filename, 'w') as f:
        f.write(ret_str)



from . import moments, dadi, moments_available, dadi_available
from .cli import SettingsStorage
from .optimizers import GlobalOptimizerAndLocalOptimizer
from .data import SFSDataHolder
from .engines import get_engine
from . import utils
import warnings
import numpy as np


def load_data_from_dir(dirname, engine, projections=None,
                       population_labels=None, outgroup=None):
    """
    Load data of SFS type from the directory. Data is considered to be
    very consistent: for example, it could be bootstrap of one dataset. All
    data should have the same projections, pop labels and so on.

    :param dirname: Path to the directory with data.
    :param engine: Engine id for data loading. Could be one of the following:
                   - dadi
                   - moments
    :param projections: Sample size of data. If None it will be chosen
                        automatically.
    :param population_labels: Labels of populations in the data.
    :param outgroup: If True then there is outgroup represented in files. Then
                     unfolded SFS will be loaded if SFS is needed.
    """
    settings = SettingsStorage()
    settings.directory_with_bootstrap = dirname
    settings.engine = engine
    settings.data_holder = SFSDataHolder(None,
                                         projections=projections,
                                         outgroup=outgroup,
                                         population_labels=population_labels)
    return settings.read_bootstrap_data()


def get_claic_score(func_ex, all_boot, p0, data, engine=None, args=(),
                    eps=1e-2, pts=None):
    r"""
    Returns CLAIC score for demographic model with specified value of `eps`.

    :param func_ex: Custom function to evaluate demographic model.
                    Usually it is model_func function from generated code of
                    GADMA. It is run by calling func_ex(p, ns, *args), where
                    p is values of parameters and ns - sample sizes.
    :param all_boot: List of bootstrapped data for CLAIC evaluation.
    :param p0: Values of parameters for ``func_ex`` demographic model.
    :param data: Original data for CLAIC evaluation. It is data that was used
                 for demographic inference.
    :param engine: Engine id for likelihood evaluations. Could be one of the
                   following:
                   - dadi
                   - moments
    :param args: Arguments of ``func_ex`` function.
    :param eps: Step size for Hessian and gradient calculations. Usually is
                between 1e-5 and 1e-2. The smaller eps is the more accurate
                CLAIC value is.
    :param pts: Deprecated parameter from GADMA version 1. If is set then
                warning is printed.

    returns: None if failed to get CLAIC due to singular matrix of Hessian.\
             Could be solved by increasing value of ``eps``.

    note: There differencies between GADMA v1 and GADMA v2, there is some\
          backward compatibility, but sometimes errors could be raised.
    """
    # The following code exists because of backward compatibility with first
    # version of GADMA, where parameters were:
    warning_msg = ("It looks like get_claic_score function is used from GADMA "
                   " of version 1.")
    if pts is not None and engine is None:
        engine = 'dadi'
        warnings.warn(warning_msg + " Deprecated argument pts is used - dadi "
                      "engine is chosen.")
    if (engine is None or
            (isinstance(engine, (list, np.ndarray)) and len(engine) == 3)):
        if engine is not None:
            pts = engine
        engine = 'moments'
        if pts is not None:
            engine = 'dadi'
        warnings.warn(warning_msg + " The 5th argument (engine in GADMA v2 vs"
                      " pts in GADMA v1) argument is not specified (None). If"
                      " some other error will happen next then please specify"
                      f" engine. Engine: {engine}, pts: {pts}, eps: {eps}")

    if not isinstance(args, tuple) and isinstance(args, float):
        eps = args
        args = ()
    # End of backward compatibility of versions.

    settings = SettingsStorage()
    settings.engine = engine
    if len(args) > 0:
        func_ex = utils.fix_args(func_ex, args)
    settings.model_func = func_ex
    engine_obj = get_engine(engine)
    engine_obj.set_data(data)
    engine_obj.set_model(settings.get_model())
    if pts is not None:
        settings.pts = pts
    return utils.get_claic_score(engine_obj, p0, all_boot,
                                 settings.get_engine_args())


def optimize_ga(data, model_func, engine, args=(),
                lower_bound=None, upper_bound=None, p_ids=None,
                X_init=None, Y_init=None, num_init=50,
                gen_size=10, mut_strength=0.2, const_mut_strength=1.1,
                mut_rate=0.2, const_mut_rate=1.2, eps=1e-2, n_stuck_gen=100,
                n_elitism=2, p_mutation=0.3, p_crossover=0.3, p_random=0.2,
                ga_maxiter=None, ga_maxeval=None,
                local_optimizer='BFGS_log', ls_maxiter=None, ls_maxeval=None,
                verbose=1, callback=None,
                save_file=None, eval_file=None, report_file=None):
    r"""
    Runs genetic algorithm optimizer in order to find best values of
    parameters for ``model_func`` demographic model from ``data``.

    :param data: Data for demographic inference.
    :param model_func: Function to use for demographic inference that
                       simulates SFS to compare it with ``data`` with
                       log-likelihood. Is called by model_func(p, ns, *args),
                       where p is values of parameters, ns - sample size and
                       args - other arguments.
    :param engine: Engine id for demographic inference. Could be one of the
                   following:
                   - 'dadi'
                   - 'moments'
    :param args: Arguments for ``model_func`` function. It is `pts` for
                 `dadi` engine and could be `dt_fac` (or nothing) for
                 `moments` engine.
    :param lower_bound: Lower bound for each demographic parameter.
    :type lower_bound: list
    :param upper_bound: Upper bound for each demographic parameter.
    :type upper_bound: list
    :param p_ids: Parameter identifiers for demographic parameters. Each
                  identifier should start with one of the following letters:
                  - n or N for size of populations;
                  - t or T for time;
                  - m or M for migration rates;
                  - s or S for fractional parameters (between 0 and 1).

                  For example valid identifiers are:
                  ['nu1F', 'nu2B', 'nu2F', 'm', 'Tp', 'T']
    :type p_ids: list
    :param X_init: list of initial example parameters. GA will be initialized
                   by those values. It could be used for combinations of
                   optimizations or for restart.
    :type X_init: list of lists
    :param Y_init: value of log-likelihood for values in X_init.
    :type Y_init: list
    :param num_init: Number of initial points to start Genetic algorithm.
    :type num_init: int
    :param gen_size: Size of generation of genetic algorithm. That is number
                     of individuals/solutions on each step of GA.
    :type gen_size: int
    :param mut_strength: Mean fraction of parameters for mutation in GA.
    :type mut_strength: float
    :param const_mut_strength: Const to change ``mut_strength`` during
                               GA according to one-fifth rule.
    :type const_mut_strength: float
    :param mut_rate: Mean rate of mutation in GA.
    :type mut_rate: float
    :param const_mut_rate: Const to change ``mut_rate`` during GA.
    :type const_mut_rate: float
    :param eps: const for model's log likelihood compare.
                Model is better if its log likelihood is greater than
                log likelihood of another model by epsilon.
    :type eps: float
    :param n_stuck_gen: Number of iterations for GA stopping: GA stops when
                        it can't improve model during n_stuck_gen generations.
    :type n_stuck_gen: int
    :param n_elitism: Number of best models from previous generation in GA
                      that will be taken to new iteration.
    :type n_elitism: int
    :param p_mutation: probability of mutation in one generation of GA.
    :type p_mutation: float
    :param p_crossover: probability of crossover in one generation of GA.
    :type p_crossover: float
    :param p_random: Probability of random generated individual in one
                     generation of GA.
    :type p_random: float
    :param ga_maxiter: Maximum number of generations in GA.
    :type ga_maxiter: int
    :param ga_maxeval: Maximum number of function evaluations in GA.
    :type ga_maxeval: int
    :param local_optimizer: Local optimizer name to run for best solution of
                            GA. Could be None or one of:
                            * 'BFGS'
                            * 'BFGS_log'
                            * 'L-BFGS-B'
                            * 'L-BFGS-B_log'
                            * 'Powell'
                            * 'Powell_log'
                            * 'Nelder-Mead'
                            * 'Nelder-Mead_log'
    :type local_optimizer: str
    :param ls_maxiter: Maximum number of iterations in local optimization.
    :type ls_maxiter: int
    :param ls_maxeval: Maximum number of function evaluations in local
                       optimization.
    :type ls_maxeval: int
    :param verbose: Verbose of output.
    :type verbose: int
    :param callback: callback to call during optimizations. (callback(x, y))
    :param save_file: File for save GA's state on current generation.
    :type save_file: str
    :param eval_file: File to save all evaluations during GA and local
                      optimization.
    :param report_file: File to write reports of GA and local optimization.
    """
    settings = SettingsStorage()
    settings._inner_data = data
    settings.model_func = model_func
    settings.engine = engine
    settings.get_engine_args = args
    settings.lower_bound = lower_bound
    settings.upper_bound = upper_bound
    settings.parameter_identifiers = p_ids
    settings.size_of_generation = gen_size
    settings.mean_mutation_rate = mut_rate
    settings.const_for_mutation_rate = const_mut_rate
    settings.mean_mutation_strength = mut_strength
    settings.const_for_mutation_strength = const_mut_strength
    settings.eps = eps
    settings.stuck_generation_number = n_stuck_gen
    settings.n_elitism = n_elitism
    settings.p_mutation = p_mutation
    settings.p_crossover = p_crossover
    settings.p_random = p_random
    settings.local_optimizer = local_optimizer

    # We could use CoreRun here but instead we will launch everything manually
    engine = get_engine(engine)
    model = settings.get_model()
    engine.set_model(model)
    engine.set_data(data)

    f = engine.evaluate
    variables = model.variables

    global_optimizer = settings.get_global_optimizer()
    local_optimizer = settings.get_local_optimizer()

    opt = GlobalOptimizerAndLocalOptimizer(global_optimizer,
                                           local_optimizer)
    result = opt.optimize(f, variables, args=args, global_num_init=num_init,
                          X_init=X_init, Y_init=Y_init, local_options={},
                          global_maxiter=ga_maxiter, local_maxiter=ls_maxiter,
                          global_maxeval=ga_maxeval, local_maxeval=ls_maxeval,
                          verbose=verbose, callback=callback,
                          eval_file=eval_file, report_file=report_file,
                          save_file=save_file)
    return result



from ..utils import check_file_existence, ensure_file_existence
import functools


class DataHolder(object):
    """
    Class for data holding.

    : param filename: name of file with data
    : param outgroup: information if there is outgroup in data
    : type outgroup: bool
    : params pop_labels: labels of populations in data
    : param seq_len: length of sequence that was used to build data
    """
    def __init__(self, filename, projections,
                 outgroup, population_labels, sequence_length):
        self.data = None
        self.filename = filename
        self.projections = projections
        self.outgroup = outgroup
        self.population_labels = population_labels
        self.sequence_length = sequence_length

        if self.filename is not None and check_file_existence(self.filename):
            self.filename = ensure_file_existence(self.filename)


class SFSDataHolder(DataHolder):
    """
    Class for SFS data holding.
    if any parameter is None then it will be taken from the file
    """
    def __init__(self, sfs_file, projections=None, outgroup=None,
                 population_labels=None, sequence_length=None):
        super(SFSDataHolder, self).__init__(sfs_file, projections,
                                            outgroup, population_labels,
                                            sequence_length)


class VCFDataHolder(DataHolder):
    """
    Class for VCF data holding.
    """
    def __init__(self, vcf_file, popmap_file, sample_sizes, outgroup,
                 population_labels=None, seq_len=None,  bed_file=None):
        if population_labels is None:
            population_labels = set()
            with open(popmap_file) as f:
                for line in f:
                    population_labels.add(line.split()[-1])
        super(VCFDataHolder, self).__init__(vcf_file, sample_sizes, outgroup,
                                            population_labels, seq_len)
        self.popmap_file = popmap_file
        self.bed_file = bed_file



from .optimizer import Optimizer, ConstrainedOptimizer
from ..utils import ContinuousVariable

import copy
import scipy
import numpy as np


_registered_global_optimizers = {}


class GlobalOptimizer(Optimizer):
    """
    Base class for global optimization.
    See :class:`gadma.optimizers.Optimizer` for more information.
    """
    def __init__(self, log_transform=False, maximize=False):
        super(GlobalOptimizer, self).__init__(log_transform, maximize)
        self.X = list()
        self.Y = list()

    def randomize(self, variables, random_type='resample',
                  custom_rand_gen=None):
        """
        Generate random solution. The type of generation could be set
        to one of three operators:

        * 'uniform' - uniform over domain.

        * 'resample' - call :meth:`resample` method for all variables.

        * 'custom' - sample values of parameters from `custom_rand_gen`.
        """
        if random_type == 'uniform':
            arr = []
            for var in variables:
                if isinstance(var, ContinuousVariable):
                    arr.append(np.random.uniform(*var.domain))
                else:
                    arr.append(np.random.choice(var.domain))
            return np.array(arr, dtype=object)
        elif random_type == 'resample':
            arr = [var.resample() for var in variables]
            return np.array(arr, dtype=object)
        elif random_type == 'custom':
            return custom_rand_gen(variables)
        else:
            raise ValueError(f"Unknown type of generation of random "
                             f"solution: {random_type}.")

    def initial_design(self, f, variables, num_init,
                       X_init=None, Y_init=None,
                       random_type='resample', custom_rand_gen=None):
        """
        Performs initial design for optimization.

        :param f: function to use for evaluations. Note that it should be
                  without arguments. Use :meth:`self.fix_f_and_args` to
                  get such function from another one with arguments.
        :param variables: variables of function. They are used for random
                          generation of their values.
        :param num_init: number of initial solutions.
        :param X_init: list of some initial solutions.
        :param Y_init: list of function values on the initial solutions.

        :returns: pair of lists X and Y. Initial points and value of fitness\
                  function on them.
        """
        X = list()
        Y = list()
        if X_init is not None:
            X = X_init
            if Y_init is not None:
                Y = Y_init
            else:
                Y = list()
            for x in X_init[len(Y_init):]:
                X.append(x)
                Y.append(f(x))
        for _ in range(num_init - len(X)):
            x = self.randomize(variables, random_type, custom_rand_gen)
            X.append(x)
            Y.append(f(x))
        return X, Y

    def optimize(self, f, variables, num_init, X_init=None, Y_init=None,
                 args=(), options={}, linear_constrain=None, maxiter=None):
        """
        Runs optimization.

        :param f: Target function to optimize.
        :type f: func
        :param variables: Variables of `f` which values should be optimized.
        :type variables: :class:`gadma.utils.VariablePool`
        :param num_init: Number of points in initial design.
        :type num_init: int
        :param X_init: List of initial points.
        :type X_init: list
        :param Y_init: Values of target function on initial points.
        :type Y_init: list
        :param args: Additional arguments of target function.
        :type args: tuple
        :param options: Additional options kwargs for optimization.
        :type options: dict
        :param linear_constrain: Linear constrain on optimized variables.
        :type linear_constrain: :class:`gadma.optimizers.LinearConstrain`
        :param maxiter: Maximum number of iterations to run.
        :type maxiter: int

        """
        raise NotImplementedError


def register_global_optimizer(id, optimizer):
    """
    Registers the specified global optimizer.
    """
    if id in _registered_global_optimizers:
        raise ValueError(f"Optimizer '{id}' is already registered.")
    if not issubclass(optimizer, GlobalOptimizer):
        raise ValueError("Optimizer is not global.")
    _registered_global_optimizers[id] = optimizer
    optimizer.id = id


def get_global_optimizer(id):
    """
    Returns the global optimizer with the specified id.
    """
    if id not in _registered_global_optimizers:
        raise ValueError(f"Optimizer '{id}' is not registered")
    return _registered_global_optimizers[id]()


def all_global_optimizers():
    """
    Returns an iterator over all registered global optimizers.
    """
    for optim in _registered_global_optimizers.values():
        yield optim()



from ..models import CustomDemographicModel, Epoch, Split, BinaryOperation
from ..utils import Variable, DiscreteVariable, DynamicVariable
import sys

FUNCTION_NAME = 'model_func'


def _print_dadi_func(model, values):
    """
    Returns string with function of demographic model for :py:mod:`dadi`.
    Parameter `values` is needed to fix dynamics of populations.

    :param model: Demographic model.
    :type model: :class:`gadma.models.model.Model`
    :param values: List of values for parameters of model.
    """
    from ..engines import DadiEngine  # to avoid cross import
    if isinstance(model, CustomDemographicModel):
        ret_str = "import importlib.util\n\n"
        ret_str += "spec = importlib.util.spec_from_file_location('module',"\
                   f" '{sys.modules[model.function.__module__].__file__}')\n"
        ret_str += "module = importlib.util.module_from_spec(spec)\n"
        ret_str += "spec.loader.exec_module(module)\n"
        ret_str += f"{FUNCTION_NAME} = module.model_func\n\n"
        return ret_str

    var2value = model.var2value(values)  # OrderedDict

    f_vars = [x for x in var2value if not isinstance(x, DiscreteVariable)]
    ret_str = f"def {FUNCTION_NAME}(params, ns, pts):\n"
    ret_str += "\t%s = params\n" % ", ".join([x.name for x in f_vars])
    ret_str += "\txx = dadi.Numerics.default_grid(pts)\n"\
               "\tphi = dadi.PhiManip.phi_1D(xx)\n"
    for ind, event in enumerate(model.events):
        if event.__class__ is Epoch:
            if event.dyn_args is not None:
                for i in range(event.n_pop):
                    dyn_arg = event.dyn_args[i]
                    value = var2value.get(dyn_arg, dyn_arg)
                    if value != 'Sud':
                        func = DynamicVariable._help_dict[value]
                        y1 = event.init_size_args[i]
                        y2 = event.size_args[i]
                        x_diff = event.time_arg
                        if isinstance(y1, Variable):
                            y1 = y1.name
                        elif isinstance(y1, BinaryOperation):
                            y1 = f"({y1.name})"
                        if isinstance(y2, Variable):
                            y2 = y2.name
                        elif isinstance(y2, BinaryOperation):
                            y2 = f"({y2.name})"
                        if isinstance(x_diff, Variable):
                            x_diff = x_diff.name
                        elif isinstance(x_diff, BinaryOperation):
                            x_diff = f"({x_diff.name})"
                        funcstr = func.func_str(y1, y2, x_diff)
                        ret_str += "\tnu%d_func = %s\n" % (i+1, funcstr)
            kwargs_with_vars = DadiEngine._get_kwargs(event, var2value)
            str_kwargs = ["%s=%s" % (k, v.name
                                     if isinstance(v, (Variable,
                                                       BinaryOperation))
                                     else v)
                          for k, v in kwargs_with_vars.items()]
            if event.n_pop == 1:
                func = 'dadi.Integration.one_pop'
            if event.n_pop == 2:
                func = 'dadi.Integration.two_pops'
            if event.n_pop == 3:
                func = 'dadi.Integration.three_pops'
            ret_str += "\tphi = %s(phi, xx, %s)\n" % (func,
                                                      ', '.join(str_kwargs))
        elif isinstance(event, Split):
            if event.n_pop == 1:
                ret_str += "\tphi = dadi.PhiManip.phi_1D_to_2D"
            else:
                ret_str += "\tphi = dadi.PhiManip.phi_%dD_to_%dD_split_%d" %\
                           (event.n_pop, event.n_pop+1, event.pop_to_div + 1)
            ret_str += "(xx, phi)\n"
    ret_str += "\tsfs = dadi.Spectrum.from_phi(phi, ns, [xx]*len(ns))\n"
    ret_str += "\treturn sfs\n"
    return ret_str


def _is_fs_via_dadi(data_holder):
    """
    Check that data is allele frequency spectrum saved for dadi.

    :param data_holder: Data holder with data.
    :type data_holder: :class:`gadma.data.data_holder.DataHolder`
    """
    import dadi
    try:
        data = dadi.Spectrum.from_file(data_holder.filename)
        return data
    except Exception as e:
        return None


def _print_load_data(data_holder, is_fs, mode='dadi'):
    """
    Returns strings for code that loads data for analysis.

    :param data_holder: Data holder with data.
    :param is_fs: If True then data isin AFS format of dadi.
    :param mode: Which mode to use, could be 'dadi' or 'moments'.
    """
    ret_str = ""
    fname = data_holder.filename
    if is_fs is None:
        ret_str += f"dd = {mode}.Misc.make_data_dict('{fname}')\n"
        ret_str += f"data = {mode}.Spectrum.from_data_dict"
        if (data_holder.population_labels is None or
                data_holder.projections is None or
                data_holder.outgroup is None):
            raise ValueError("Please specify data_holder carefully. "
                             "Some attribute is None.")
        lbls = list(data_holder.population_labels)
        size = list(data_holder.projections)
        outg = data_holder.outgroup
        ret_str += f"(dd, {lbls}, {size}, polarized={outg})\n"
    else:
        data = is_fs
        ret_str += f"data = {mode}.Spectrum.from_file('{fname}')\n"
        if (data_holder.projections is not None and
                list(data.sample_sizes) != list(data_holder.projections)):
            size = data_holder.projections
            ret_str += "data = data.project(%s)\n" % (str(list(size)))
        if data.folded == data_holder.outgroup:
            ret_str += "data = data.fold()\n"
        if data.pop_ids is None:
            if data_holder.population_labels is None:
                data.pop_ids = [f"Pop{i}" for i in range(data.ndim)]
            else:
                data.pop_ids = data_holder.population_labels
            ret_str += f"data.pop_ids = {data.pop_ids}\n"
        if (data_holder.population_labels is not None and
                list(data.pop_ids) != list(data_holder.population_labels)):
            d = {x: i for i, x in enumerate(data.pop_ids)}
            d = [d[x] for x in data_holder.population_labels]
            ret_str += "data = np.transpose(data, %s)\n" % (str(d))
            ret_str += f"data.pop_ids = {data_holder.population_labels}\n"
    return ret_str


def _print_dadi_load_data(data_holder):
    """
    Check :func:`_print_load_data` for more information.
    Checks that `data_holder` is in AFS format of dadi and runs
    :func:`_print_load_data` with 'dadi' mode.
    """
    is_fs = _is_fs_via_dadi(data_holder)
    return _print_load_data(data_holder, is_fs, mode='dadi')


def _print_p0(values):
    """
    Returns code for p0 with values.

    :param values: Values for p0.
    """
    return "p0 = [%s]\n" % ", ".join([str(x) for x in values])


def _print_dadi_simulation():
    """
    Returns string of code about simulation with dadi.
    """
    ret_str = f"func_ex = dadi.Numerics.make_extrap_log_func"\
              f"({FUNCTION_NAME})\n"
    ret_str += "model = func_ex(p0, ns, pts)\n"
    return ret_str


def _print_main(engine, values, mode='dadi', nanc=None):
    """
    Returns string for main part of generated code.

    :param engine: Engine that was used with data and model.
    :param values: best values of model parameters.
    :param mode: 'dadi' or 'moments'
    :param nanc: Size of ancestral population. It could be known if inference
                 was made by other engine and nanc from this engine will be
                 different to optimal.
    """
    ret_str = f"ll_model = {mode}.Inference.ll_multinom(model, data)\n"
    ret_str += "print('Model log likelihood (LL(model, data)): "\
               "{0}'.format(ll_model))\n"
    ret_str += f"\ntheta = {mode}.Inference.optimal_sfs_scaling"\
               "(model, data)\n"

    ret_str += "print('Optimal value of theta: {0}'.format(theta))\n"

    if nanc is not None:
        ret_str += f"Nanc = {nanc}  # {mode} was not used for inference\n"
        return ret_str

    mu_is_val = engine.model.mu is not None
    data_holder_is_val = engine.data_holder is not None
    seq_len_is_val = engine.data_holder.sequence_length is not None

    mu_and_L = mu_is_val and data_holder_is_val and not seq_len_is_val

    if engine.model.theta0 is not None or mu_and_L:
        if engine.model.theta0 is not None:
            ret_str += f"theta0 = {engine.model.theta0}\n"
            ret_str += "Nanc = int(theta / theta0)\n"
        elif mu_and_L:
            ret_str += f"mu = {engine.model.mu}\n"
            ret_str += f"L = {engine.data_holder.sequence_length}\n"
            ret_str += "Nanc = int(theta / (4 * mu * L))\n"
        ret_str += "print('Size of ancestral population: {0}'.format(Nanc))\n"
    else:
        ret_str += "Nanc = None\n"
    return ret_str


def _print_dadi_main(engine, values, nanc=None):
    ret_str = _print_p0(values)
    ret_str += _print_dadi_simulation()
    ret_str += _print_main(engine, values, mode='dadi', nanc=nanc)
    return ret_str


def print_dadi_code(engine, values, pts, filename,
                    nanc=None, gen_time=None, gen_time_units=None):
    """
    Generates code for `dadi` to file. Code have function of demographic model
    that simulates AFS and main part where simulation takes place as well as
    calculation of log-likelihood.

    :param engine: Engine that was used with data and model.
    :param values: Value of model parameters.
    :param pts: Grid sizes for dadi.
    :param filename: File to save generated code.
    :param nanc: Size of ancestral population. Is used when other engine was
                 used for inference.
    :param gen_time: Time of one generation in units of ``gen_time_units``.
    :param gen_time_units: Units of time. String.

    :note: the last two arguments are ignored as dadi could not draw models.
    """
    ret_str = "import dadi\nimport numpy as np\n\n"
    ret_str += _print_dadi_func(engine.model, values)
    ret_str += "\n"
    ret_str += _print_dadi_load_data(engine.data_holder)
    ret_str += "pts = %s\nns = data.sample_sizes\n\n" % str(pts)

    values = [val for var, val in engine.model.var2value(values).items()
              if not isinstance(var, DiscreteVariable)]
    ret_str += _print_dadi_main(engine, values, nanc)
    if filename is None:
        return ret_str
    with open(filename, 'w') as f:
        f.write(ret_str)
